---------------------------------------------------------------------------
-- FileName     : sram_8_model.vhd
-- Function     : 8bit_data SRAM Model for simulation
-- Author       : F.O (ProXi)
-- Date         : 2018/11/07
-- Company      : ProXi
-- ylz     : ڍׂ  http://www.proxi.co.jp/technolo/pci_bus_test_bench.htm Q
--              : {\[X ASIANL@ (Attributed SIgnAl Naming)  ŋLqĂB
--              :   ڍׂ http://www.proxi.co.jp/technolo/asian.htm Q
---------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
---------------------------------------------------------------------------

---------------------------------------------------------------------------
-- 8bit_data SRAM Model for simulation
---------------------------------------------------------------------------
entity sram_8_model is
    generic ( RAM_ADR_WIDTH : integer   := 4;       -- ram address width
              SET_HIZ_VAL   : std_logic := 'H';     -- select one from 'Z','H','L'
              t_OE          : time      := 20.0 ns  -- time bitween output enable to output valid
            );
    port (
      -- in
        a_cs_l      : in  std_logic;
        a_rd_l      : in  std_logic;
        a_wr_l      : in  std_logic;
        av_adr      : in  std_logic_vector(RAM_ADR_WIDTH-1 downto 0);
      -- inout
        av_data     : inout std_logic_vector(7 downto 0)
    );
end sram_8_model;

architecture rtl of sram_8_model is
    subtype RAM_DATA is std_logic_vector(7 downto 0);
    type RAM_ARRAY is array (0 to 2**RAM_ADR_WIDTH -1) of RAM_DATA;

    signal xav_ram_data     : RAM_ARRAY;
    signal xi_ram_adr       : integer range 0 to 2**RAM_ADR_WIDTH -1;
    signal xa_cs            : std_logic;
    signal xa_rd            : std_logic;
    signal xa_wr            : std_logic;
 
begin   
    xa_cs <= not a_cs_l;
    xa_rd <= not a_rd_l;
    xa_wr <= not a_wr_l;

    xi_ram_adr <= conv_integer(av_adr);         -- address to ram_element_no

    -- ram write
    process (xa_wr) begin
        if (xa_wr'event and xa_wr='0') then
            if ( (xa_cs='1') and (xa_rd='0') )  then
                xav_ram_data(xi_ram_adr)(7 downto 0) <= av_data(7 downto 0);
            end if;
        end if;
    end process;

    -- ram read
    av_data(7 downto 0) <= xav_ram_data(xi_ram_adr)(7 downto 0)  after t_OE
        when ( (xa_cs='1') and (xa_rd='1') and (xa_wr='0') ) else (others => SET_HIZ_VAL);

end rtl;

